/******************** dataDisplay.c *******************************/
/*                                                                */
/*                 VIRGO DATA DISPLAY                             */
/*                                                                */
/*   MODULE     : dataDisplay.c                                   */
/*   AUTHOR     : D. Verkindt (LAPP,Annecy)                       */
/*                verkindt@lapp.in2p3.fr                          */
/*   VERSION    : v9r12p19                                        */
/*   DATE       : 17 Oct 2014                                     */
/*                                                                */
/*   This application is a Data Display for the Virgo             */
/*   Gravitational Waves Interferometric Detector.                */
/*   It reads data under the frame format, either from files      */
/*   or from ethernet (using Cm encapsulation)                    */
/*   It uses the GUI Xforms and the Root library for plots,       */
/*   it allows to dump data, to do various plot types             */
/*   and, under UNIX alpha workstations, to hear the data stream. */
/*                                                                */
/* Copyright (C) 2003, D. Verkindt                                */
/* For the licensing terms see the LICENSE file.                  */
/*                                                                */
/******************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <fcntl.h> 
#include <stdarg.h>
#include <dirent.h>
#include <assert.h>
#include <unistd.h>
#include <locale.h>
#include "UJobL.h"
#include "dataDisplay.h"
#include "URootAccess.h"
#ifdef MEMWATCH
#    include "memwatch.h"
#endif

#include "RVersion.h"


struct UJstructH *UserUJstructH;
struct DyTool* Dy;                  /* pointer to the main dataDisplay structure            */
struct DyPlot* Plot[NPLOT_MAX];     /* array of pointers to plot structures                 */
struct FrVect *UFrCurVect = NULL;   /* last vector of data read                             */
struct FrameH *UFrCurFrameH = NULL; /* last frame read                                      */
int UFrCurFrameHSize;               /* size in bbytes of the last frame read                */
struct soundhdr wavehdr;            /* structure to contains wave file header               */
struct timeval dytv;                /* timer to know how much time was passed in a function */
double         dyptime;             /* previous result of the timer                         */
char           dytimestr[256];      /* string with current time and dtime                   */
double ptime[DBGTIME_LEVELMAX+1];
char fctnamep[DBGTIME_LEVELMAX+1][256];

#ifdef USE_VEGA
struct DyExtern* DyExternSource;
#endif

FD_mainbrowser *mainbrowser;
FD_inputfile   *inputfile;
FD_mainpanel   *mainpanel;
FD_userfuncbrowser *userfuncbrowser;
FD_plotaudio   *plotaudio;
FD_plotgenepar *plotgenepar;
FD_plot1d      *plot1d;
FD_plot2d      *plot2d;
FD_plotraw     *plotraw;
FD_plottime    *plottime;
FD_plotfft     *plotfft;
FD_plotffttime *plotffttime;
FD_plot1dtime  *plot1dtime;
FD_plotrawtime *plotrawtime;
FD_plottrfct   *plottrfct;
FD_dumpdata    *dumpdata;
FD_getdd       *getdd;
FL_FORM        *helppanel;
FL_FORM        *coffeepanel;
FL_FORM        *userprochelp;
FD_cmservers   *cmservers;
FD_getshm      *getshm;
FD_dydebug     *dydebug;
FD_dyoptions   *dyoptions;
FD_dyoutput    *dyoutput;
FD_reffile     *reffile;

/***  unused  ***/
static char DdHostAlpha[LSTRING_MAX];
static char DdHostLynxOS[LSTRING_MAX];


/*------------------------------------------------------ Main ---*/

#ifndef USE_VEGA
int main (int argc, char *argv[])
#else
int DyMain (struct DyExtern* dySource)
#endif
{
 int i;
 int myargc=1;
 char *buffer, date_heure[24];
 time_t tt;

#ifdef MEMWATCH
 mwStatistics(2);
#endif

#ifndef USE_VEGA
 printf("\n\n");
 printf("-----------------------------------------------------------------------------\n");
 printf("                         VIRGO Data Display (%s) \n", DYVERSION);
 printf("-----------------------------------------------------------------------------\n");
 tt = time(NULL);
 strftime(date_heure,24,"%b %d %Y  %T",localtime(&tt));
 printf("Started on:   %s  \n", date_heure); 
 printf("Compiled on:  %s  %s   Source: %s \n\n",__DATE__,__TIME__,__FILE__); 
 printf(" using:\n");
 FrLibVersion(stdout);
 printf("  ROOT %s\n", ROOT_RELEASE);
 printf("  GUI Xforms (v%dr%d)\n", FL_VERSION, FL_REVISION);
#ifndef NOCM
 printf("  Cm   (%s)\n", CMVERSION);
 printf("  DyFd (%s)\n", DYFDVERSION);
 printf("Contact: %s\n", CURRENT_MANAGER);
#endif
 printf("-----------------------------------------------------------------------------\n");
#else
 printf("\n\n");
 printf("-----------------------------------------------------------------------------\n");
 printf("                         Viewer based on dataDisplay (%s) \n", DYVERSION);
 printf(" using GUI Xforms (v%dr%d)\n", FL_VERSION, FL_REVISION);
 printf("-----------------------------------------------------------------------------\n");
#endif
 printf("\n Usage: dataDisplay [config] -f [start GPS] -l [duration] -o [filename] -t [title] -b -q -h\n");
 printf("\n");

 /*--- Init the timer for debug ---*/
 DyDebugTime();

 /*--- Init Dy structure ---*/
 Dy = (struct DyTool *)calloc(1,sizeof(struct DyTool));
 Dy->AutoConf          = NULL;
 Dy->AutoConfFlag      = OFF;
 Dy->startTime         = -1;
 Dy->duration          = -1;
 Dy->startTimeCmd      = -1;
 Dy->durationCmd       = -1;
 Dy->plotsNotDisplayed = OFF;
 Dy->BgRun             = OFF;
 Dy->QuitAtStop        = OFF;
 Dy->outPicture        = NULL;
 Dy->titleCmd          = NULL;

 /*--- Fill the plots and panels type names ---*/
 DyMakeTypeNames();


#ifndef USE_VEGA

 /*--- Get configuration file if provided ---*/

 if (argc>1 && strncmp(argv[1],"-",1) != 0)
    {if (is_a_file(argv[1]) == 0) {printf("***!! %s not found !!***\n\n", argv[1]); exit(0);}
     Dy->AutoConf = my_strcpy(Dy->AutoConf,argv[1]);
     Dy->AutoConfFlag = ON;
    }

 /*--- Get the arguments passed to data Display ---*/

 for (i=1; i<argc; i++)
    {
     if (strncmp(argv[i],"-f",2) == 0 && i+1<=argc)          {Dy->startTimeCmd = atoi(argv[i+1]); Dy->startTime = atoi(argv[i+1]); i++;}
     else if (strncmp(argv[i],"-l",2) == 0 && i+1<=argc)     {Dy->durationCmd = atol(argv[i+1]); Dy->duration = atol(argv[i+1]); i++;}
     else if (strncmp(argv[i],"-b",2) == 0)                   Dy->BgRun = ON;
     else if (strncmp(argv[i],"-q",2) == 0)                   Dy->QuitAtStop = ON;
     else if (strncmp(argv[i],"-o",2) == 0 && i+1<=argc)     {Dy->outPicture = my_strcpy(Dy->outPicture, argv[i+1]); i++;}
     else if (strncmp(argv[i],"-t",2) == 0 && i+1<=argc)     {Dy->titleCmd   = my_strcpy(Dy->titleCmd, argv[i+1]); i++;}
     else if (strncmp(argv[i],"-noplot",7) == 0)              Dy->plotsNotDisplayed = ON;
     else if (strncmp(argv[i],"-debug",6) == 0 && i+1<=argc) {Dy->debugType = 0xffff; Dy->debug = atoi(argv[i+1]); i++;}
     else if (strncmp(argv[i],"-h",2) == 0)
        {printf(" [config]        (default=NULL)         = name of configuration file\n");
         printf("  -f [start GPS] (default=now)          = start GPS time \n");
         printf("  -l [duration]  (default=LONG_MAX)     = number of seconds of data to read \n");
         printf("  -b                                    = run in background\n");
         printf("  -q                                    = quit when dataDisplay stops\n");
         printf("  -o [filename]  (default=config)       = output picture (.jpg, .root)\n");
         printf("  -t [title]     (default=current date) = a title for the canvas of plots\n");
         printf("  -noplot        (default=0)            = plots not displayed but dd.root file produced\n");
         printf("  -debug [level] (default=0)            = run with debug level\n");
         printf("\n");
         exit(0);
        }
    }

 if (Dy->BgRun == ON) Dy->QuitAtStop = ON;

 printf("Argument passed: AutoConf=%s  startTime=%d  duration=%d  "
        "BgRun=%d  QuitAtStop=%d  outPicture=%s  titleCmd=%s  noplot=%d  debug=%d\n",
         Dy->AutoConf, Dy->startTime, Dy->duration,
         Dy->BgRun, Dy->QuitAtStop, Dy->outPicture, Dy->titleCmd, Dy->plotsNotDisplayed, Dy->debug);
#endif


 RA_InitROOT(); /* Init the ROOT environment */
 signal (SIGINT, quitHandler);      /*--- Take care only of Ctrl-C ---*/

#ifndef USE_VEGA
 DyXformIni (argv, &myargc);
 DyIni (argv, argc);
#else
 DyXformIni (0, &myargc);
 DyExternSource = dySource; /* transfer the fr channel pointer */
 DyIni (0,0);
#endif

 /*--- Set the debug level of the frame library ---*/
 FrLibSetLvl(0);

 /*--- Set the function to call when a panel is closed ---*/
 fl_set_form_atclose(mainpanel->mainpanel, DyQuitAtClose, "mainpanel");
 fl_set_form_atclose(mainbrowser->mainbrowser, DyAtClose, "mainbrowser");
 fl_set_form_atclose(userfuncbrowser->userfuncbrowser, DyAtClose, "userfuncbrowser");
 fl_set_form_atclose(plotfft->plotfft, DyAtClose, "plotfft");
 fl_set_form_atclose(plotffttime->plotffttime, DyAtClose, "plotffttime");
 fl_set_form_atclose(plot1dtime->plot1dtime, DyAtClose, "plot1dtime");
 fl_set_form_atclose(plottrfct->plottrfct, DyAtClose, "plottrfct");
 fl_set_form_atclose(plottime->plottime, DyAtClose, "plottime");
 fl_set_form_atclose(plot1d->plot1d, DyAtClose, "plot1d");
 fl_set_form_atclose(plot2d->plot2d, DyAtClose, "plot2d");
 fl_set_form_atclose(plotaudio->plotaudio, DyAtClose, "plotaudio");
 fl_set_form_atclose(plotgenepar->plotgenepar, DyAtClose, "plotgenepar");
 fl_set_form_atclose(inputfile->inputfile, DyAtClose, "inputfile");
 fl_set_form_atclose(dumpdata->dumpdata, DyAtClose, "dumpdata");
 fl_set_form_atclose(cmservers->cmservers, DyAtClose, "cmservers");
 fl_set_form_atclose(getdd->getdd, DyAtClose, "getdd");
 fl_set_form_atclose(getshm->getshm, DyAtClose, "getshm");
 fl_set_form_atclose(plotraw->plotraw, DyAtClose, "plotraw");
 fl_set_form_atclose(plotrawtime->plotrawtime, DyAtClose, "plotrawtime");
 fl_set_form_atclose(dydebug->dydebug, DyAtClose, "dydebug");
 fl_set_form_atclose(dyoptions->dyoptions, DyAtClose, "dyoptions");
 fl_set_form_atclose(dyoutput->dyoutput, DyAtClose, "dyoutput");
 fl_set_form_atclose(reffile->reffile, DyAtClose, "reffile");

 /*--- Start automatically when a configuration file is provided on command line ---*/
 if (Dy->AutoConf != NULL)
    {
     buffer = DyReadConfigFile(Dy->AutoConf);
     if (buffer == NULL) exit(0);
     DyGetConfig(buffer);
     free(buffer);
     if (Dy->debug>10) DyXformChecks(NULL,0);
     DyXformStart(NULL,0);
    }

/***
 while (fl_do_forms());
***/

/*--- Prevent the use of french locale (which introduce 0, instead of . for decimals) ---*/
 setlocale(LC_ALL,"C");

/*--- Main loop scanning ROOT events ---*/
 RA_Main();

#ifndef USE_VEGA
 exit(0);
#else
 return(0);
#endif
}


