/*******************************************************/
/*                                                     */
/* File: sound2frame.c                                 */
/* Convert a WAV file into a frame file                */
/* D. Verkindt (verkindt@lapp.in2p3.fr)  23 June 2008  */
/*                                                     */
/*******************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <FrameL.h>

#define ON 1
#define OFF 0
#define AUDIO_BUF_SIZE  1024

struct soundhdr {
  char  riff[4];        /* "RIFF"                                  */
  long  flength;        /* file length in bytes                    */
  char  wave[4];        /* "WAVE"                                  */
  char  fmt[4];         /* "fmt "                                  */
  long  block_size;     /* in bytes (generally 16)                 */
  short format_tag;     /* 1=PCM, 257=Mu-Law, 258=A-Law, 259=ADPCM */
  short num_chans;      /* 1=mono, 2=stereo                        */
  long  srate;          /* Sampling rate in samples per second     */
  long  bytes_per_sec;  /* bytes per second                        */
  short bytes_per_samp; /* 2=16-bit mono, 4=16-bit stereo          */
  short bits_per_samp;  /* Number of bits per sub-sample            */
  char  data[4];        /* "data"                                  */
  long  dlength;        /* data length in bytes (filelength - 44)  */
};

void DyWaveToWave (char *waveoutfilename, char *wavefilename);
void DyWaveToFrame (char *framefilename, char *wavefilename, double framedt);
void DyFrameToWave (char *wavefilename, char *framefilename, char *chname, double start, double end);
void DySoundReadHeader (char *buffer);
void DySoundFillHeader (struct soundhdr *hdr, long srate, long nBytes, short bits_per_sample);

int sound_debug;
FILE *fin, *fout;
struct FrFile *iFile, *oFile;

/*-------------------------------------------------- quitHandler -------*/
void quitHandler(int sig)
{
 if (fin)   fclose(fin);
 if (fout)  fclose(fout);
 if (iFile) FrFileIEnd(iFile);
 if (oFile) FrFileOEnd(oFile);
 return;
}

/*--------------------------------------------------- main ----*/
int main (long argc, char *argv[])
{
 char chname[256];
 FILE *fin;
 double framedt, gps_start, gps_end;
 struct FrFile *oFile;
 struct FrameH *frame;
 struct FrAdcData *adc;

 if (argc == 1)
    {printf("\n");
     printf("\033[31;1msound2frame: I need at least one argument: the input file name:\033[0m \n");
     printf("--------------------------------------------------------------- \n");
     printf("- \033[34;1mCASE 1:\033[0m If file's name ends with .gwf, I will convert a frames file to a wave audio file (frametowave.wav).\n");
     printf("    In this case, optional arguments are channnel's name, gps_start, gps_end (or duration if less than 700000000) .\n");
     printf("    \033[34;1mSyntax is for instance: sound2frame inputfile.gwf \"Pr_B1* Em_AC*01\" 8788001500 100\033[0m \n");
     printf("\n");
     printf("- \033[34;1mCASE 2:\033[0m If file's name ends not with .gwf, I will convert a wave audio file to a frames file (wavetoframe.gwf).\n");
     printf("    In this case, optional argument is the frame length in seconds.\n");
     printf("    \033[34;1mSyntax is for instance: sound2frame inputfile.wav 10\033[0m \n");
     printf("\n");
     printf("- \033[34;1mCASE 3:\033[0m If frame length is <= 0, I will convert a wave audio file to a wave audio file (wavetowave.wav).\n");
     printf("    and frame length is interpreted as a resampling frequency.\n");
     printf("    \033[34;1mSyntax is for instance: sound2frame inputfile.wav -8000\033[0m \n");
     printf("\n\n");
     exit(0);
    }

 signal(SIGINT,quitHandler);      /*--- Take care of Ctrl-C ---*/

 fin = NULL;   fout = NULL;
 oFile = NULL; iFile = NULL;

 sound_debug = 0;
 if (argc > 2 && strcmp(argv[2],"debug") == 0) sound_debug = 1;

 /*--- Convert frames file into wave audio file ---*/ 
 if (strstr(argv[1],".gwf") != NULL)
    {strcpy(chname,"fraudio");
     gps_start = 0; gps_end = 0;
     if (argc > 2 && strcmp(argv[2],"debug") != 0) strcpy(chname, argv[2]);
     if (argc > 3) gps_start = atof(argv[3]);
     if (argc > 4) gps_end = atof(argv[4]);
     DyFrameToWave("frametowave.wav",argv[1], chname, gps_start, gps_end);
    }

 /*--- Convert wave audio file into frames file ---*/
 else
    {framedt = 1;
     if (argc > 2) framedt = atof(argv[2]);
     if (framedt == 0) DyWaveToWave("wavetowave.wav",argv[1]);
     else              DyWaveToFrame("wavetoframe.gwf",argv[1], framedt);
    }

 return;
}


/*--------------------------------------------------- DyWaveToWave ----*/
void DyWaveToWave (char *waveoutfilename, char *wavefilename)
{
 unsigned char valuec;
 short value;
 struct soundhdr hdr;
 struct soundhdr newhdr;
 char *buf;
 long i, nData;
 int bigendian;
 struct FrameH *frame;
 struct FrAdcData *adc;


 fin = fopen(wavefilename,"r");
 if (fin == NULL) printf("Could not open wave file %s \n", wavefilename);

 /*--- Read the wav file header ---*/
 buf = (char *)(&hdr);
 fread(buf, 1, sizeof(hdr), fin);
 DySoundReadHeader(buf);

 if (hdr.riff[0] != 'R') {printf("DyWaveToWave> %s if not a wav file \n", wavefilename); return;}
 if (hdr.bits_per_samp != 8 && hdr.bits_per_samp != 16) {printf("DyWaveToWave> %s is not 8 or 16 bits wav file \n", wavefilename); return;}
 if (hdr.format_tag != 1 && hdr.format_tag != 259) {printf("DyWaveToWave> %s is not PCM format wav file \n", wavefilename); return;}

 if (hdr.num_chans == 256 * buf[30] + buf[31])
    {bigendian = ON;
     printf("DyWaveToWave> num_chans=%d  buf[30]=%ud  buf[31]=%ud  We are in big endian\n", hdr.num_chans, buf[30], buf[31]);
    }
 else
    {bigendian = OFF;
     printf("DyWaveToWave> num_chans=%d  buf[30]=%ud  buf[31]=%ud  We are in little endian\n", hdr.num_chans, buf[30], buf[31]);
    }

 if (sound_debug > 0) return;

 /*--- read data from wave file ---*/

 fout = fopen(waveoutfilename,"w");
 if (fout == NULL) printf("DyWaveToWave> Could not open wave file %s \n", waveoutfilename);
 
 memcpy((char *)(&newhdr), (char *)(&hdr), sizeof(hdr));
 if (hdr.num_chans == 2)
    {newhdr.dlength = hdr.dlength / 2;
     newhdr.flength = newhdr.dlength + sizeof(newhdr);
     newhdr.num_chans = 1;
     newhdr.bytes_per_sec = hdr.bytes_per_sec / 2;
     newhdr.bytes_per_samp = hdr.bytes_per_samp / 2;
    }

 fwrite(&newhdr,sizeof(newhdr),1,fout);

 nData = hdr.dlength / (hdr.bytes_per_samp / hdr.num_chans);
 printf("DyWaveToWave> %ld bytes with %d bytes per sample and %d channels --> nData=%d\n", hdr.dlength, hdr.bytes_per_samp, hdr.num_chans);

 for (i=0; i<nData; i++)
    {
     if (hdr.bits_per_samp == 16) fread(&value, sizeof(unsigned short), 1, fin);
     else                         fread(&valuec, sizeof(unsigned char), 1, fin);

     if (hdr.num_chans == 2 && i%2 == 1) continue;

     if (hdr.bits_per_samp == 16)
        {fwrite(&value, sizeof(unsigned short), 1, fout);
         if (i%100000 < 10) printf("DyWaveToWave> i=%d  value=%d  \n", i, value);
        }
     else
        {fwrite(&valuec, sizeof(unsigned char), 1, fout);
         if (i%100000 < 10) printf("DyWaveToWave> i=%d  valuec=%d  \n", i, valuec);
        }
    }

 fclose(fout);
 fclose(fin);

 return;
}


/*--------------------------------------------------- DyWaveToFrame ----*/
void DyWaveToFrame (char *framefilename, char *wavefilename, double framedt)
{
 unsigned char valuec;
 short value;
 struct soundhdr hdr;
 char *buf;
 long i;
 int is, nData, vectnData, bigendian;
 struct FrameH *frame;
 struct FrAdcData *adc;

 fin = fopen(wavefilename,"r");
 if (fin == NULL) printf("Could not open wave file %s \n", wavefilename);

 /*--- Read the wav file header ---*/
 buf = (char *)(&hdr);
 fread(buf, 1, sizeof(hdr), fin);
 DySoundReadHeader(buf);

 if (hdr.riff[0] != 'R') {printf("DyWaveToFrame> %s if not a wav file \n", wavefilename); return;}
 if (hdr.bits_per_samp != 8 && hdr.bits_per_samp != 16) {printf("DyWaveToFrame> %s is not 8 or 16 bits wav file \n", wavefilename); return;}
 if (hdr.format_tag != 1 && hdr.format_tag != 259) {printf("DyWaveToFrame> %s is not PCM format wav file \n", wavefilename); return;}

 if (hdr.num_chans == 256 * buf[30] + buf[31])
    {bigendian = ON;
     printf("DyWaveToFrame> num_chans=%d  buf[30]=%ud  buf[31]=%ud  We are in big endian\n", hdr.num_chans, buf[30], buf[31]);
    }
 else
    {bigendian = OFF;
     printf("DyWaveToFrame> num_chans=%d  buf[30]=%ud  buf[31]=%ud  We are in little endian\n", hdr.num_chans, buf[30], buf[31]);
    }

 if (sound_debug > 0) return;

 /*--- read data from wave file ---*/

 vectnData     = (int)(hdr.srate * framedt);
 oFile         = FrFileONew(framefilename,-1);
 frame         = FrameHNew("frame_from_wave");
 frame->GTimeS = time(NULL) - FRGPSOFF + FRGPSDELTA;
 frame->dt     = framedt;
 adc = FrAdcDataNew(frame,"fraudio", hdr.srate, vectnData, 16);

 nData = hdr.dlength / (hdr.bytes_per_samp / hdr.num_chans);
 printf("DyWaveToFrame> %ld bytes with %d bytes per sample (%d bits) and %d channels --> nData=%d\n",
        hdr.dlength, hdr.bytes_per_samp, hdr.bits_per_samp, hdr.num_chans, nData);

 is = 0;
 for (i=0; i<nData; i++)
    {
     if (hdr.bits_per_samp == 16) fread(&value, sizeof(unsigned short), 1, fin);
     else                         fread(&valuec, sizeof(unsigned char), 1, fin);

     if (hdr.num_chans == 2 && i%2 == 1) continue;

     if (is == vectnData) {FrameWrite(frame,oFile); frame->GTimeS += framedt; is = 0;}
     if (hdr.bits_per_samp == 16)
        {adc->data->dataS[is] = value;
         if (i%1000000 < 10) printf("DyWaveToFrame> i=%d  is=%d  value=%d  adc->data=%d\n", i, is, value, adc->data->dataS[is]);
        }
     else
        {adc->data->dataS[is] = (short)(valuec*256 - 32768);
         if (i%100000 < 10) printf("DyWaveToFrame> i=%d  is=%d  valuec=%d  adc->data=%d\n", i, is, valuec, adc->data->dataS[is]);
        }
     is++;
    }

 if (is != 1) {for (i=is; i<vectnData; i++) adc->data->dataS[i]=0; FrameWrite(frame,oFile);}
 fclose(fin);
 FrameFree(frame);
 FrFileOEnd(oFile);

 return;
}


/*--------------------------------------------------- DyFrameToWave ----*/
void DyFrameToWave (char *wavefilename, char *framefilename, char *chname, double start, double end)
{
 double value, value_norm, valueBuf, valueBuf2;
 double delta, min_value, max_value;
 int z, isamp, resampFactor;
 short outshort, previousShort, outshort0;
 unsigned short outshort_pb;
 unsigned short outshort_gb;
 unsigned char outchar;
 unsigned char outchar2;
 struct soundhdr hdr;
 long i, nBytes;
 int vectnData, iFirst;
 double framedt, file_start, file_end, tt, tStart, tEnd, srate;
 struct FrameH *frame;
 struct FrVect *vect;

 /*--- Open input frames file ---*/
 iFile = FrFileINew(framefilename);
 if (iFile == NULL) printf("DyFrameToWave> Could not open frame file %s \n", framefilename);

 /*--- read data from frame file from tStart to tEnd ---*/

 file_start = FrFileITStart(iFile);
 file_end   = FrFileITEnd(iFile);
 if (start > 0) tStart = start;
 else           tStart = file_start;
 if (end > 700000000) tEnd = end;
 else if (end > 0)    tEnd = tStart + end;
 else                 tEnd = file_end;

 tt        = tStart;
 framedt   = 1.;
 vectnData = 2000;
 iFirst    = ON;
 isamp     = 0;

 while (tt < tEnd)
    {
     /*--- Get frame ---*/
     frame = FrameReadTChnl(iFile, tt, chname);
     if (frame == NULL)
        {printf("frame %.1f is NULL\n", tt);
         if (iFirst == ON) exit(0);
         value = 0;
         for (i=0; i<vectnData; i++) fwrite(&value, sizeof(short), 1, fout);
         tt += framedt;
         continue;
        }
     else
        {framedt = frame->dt;
         printf("Read frame %d \n", frame->GTimeS);
        }

     /*--- Get vector of data ---*/
     vect = FrameFindVect(frame, chname);
     if (vect == NULL)
        {printf("vect %s of frame %.1f is NULL\n", chname, frame->GTimeS);
         if (iFirst == ON) exit(0);
         outshort = 0;
         for (i=0; i<vectnData; i++) fwrite(&outshort, sizeof(short), 1, fout);
         tt += framedt;
         continue;
        }
     else
        {vectnData = vect->nData;}

     /*--- First time, open output file and write the header ---*/
     if (iFirst == ON)
        {iFirst = OFF;
         if (vect->type != FR_VECT_2S) printf("DyFrameToWave> Channel %s does not contain short values.\n", chname);
         nBytes = 2 * vect->nData * (int)((tEnd-tStart)/frame->dt);
         srate  = 1./vect->dx[0];
         resampFactor = 1;
         if (srate < 5512.)
            {while (resampFactor * srate < 5512.) resampFactor++;
             printf("DyFrameToWave> WARNING: Sampling frequency of %.2f Hz is too low for sound playing\n",srate);
             printf("DyFrameToWave> Data will be interpolated to %.2f Hz to be in allowed frequency band: 5512 to 44100 Hz\n", srate*resampFactor);
             srate *= resampFactor;
             nBytes *= resampFactor;
            }
         DySoundFillHeader(&hdr, (long)srate, nBytes, 16);
         DySoundReadHeader((char *)(&hdr));
         if (sound_debug > 0) {FrameFree(frame); FrFileIEnd(iFile); return;}
         fout = fopen(wavefilename,"w");
         if (fout == NULL) printf("DyFrameToWave> Could not open wave file %s \n", wavefilename);
         fwrite(&hdr,sizeof(hdr),1,fout);
        }

              /*--- On lappc-vi22 (linux machine) value outshort=0x7f44 is stored as follows: first 8 bits: 44 , last 8 bits: 7f ---*/
              /**
                 outshort = vect->dataS[i];
                  memcpy(&outchar,&outshort,2); outchar2 = *(&outchar+1);
                  outshort_gb = (unsigned short)(256*outchar + outchar2);
                  outshort_pb = (unsigned short)(256*outchar2 + outchar);
                  if (i<10) printf("value=%g  outshort=%d  outshort_gb=%d  outshort_pb=%d  outchar=%d  outchar2=%d \n",
                            value, outshort, outshort_gb, outshort_pb, outchar, outchar2);
               **/

     /*--- Write the data converted into unsigned shorts ---*/
     for (i=0; i<vectnData; i++)
         {
          if (vect->type == FR_VECT_C)
             {outshort = (short)vect->data[i];}
          else if (vect->type == FR_VECT_1U)
             {outshort = (short)vect->dataU[i];}
          else if (vect->type == FR_VECT_2U)
             {outshort = (short)vect->dataUS[i];}
          else if (vect->type == FR_VECT_4U)
             {outshort = (short)vect->dataUI[i];}
          else if (vect->type == FR_VECT_4S)
             {outshort = (short)vect->dataI[i];}

          else if (vect->type == FR_VECT_2S)
             {outshort = vect->dataS[i];}

          else
             {value = FrVectGetV(vect,i);
              if (isamp == 0) {max_value = value; min_value = value;}
              if (isamp < AUDIO_BUF_SIZE)
                 {max_value = (value > max_value) ? value : max_value; 
                  min_value = (value < min_value) ? value : min_value;
                 }
              if (isamp == AUDIO_BUF_SIZE)
                 {delta = (max_value - min_value);
                  min_value -= 0.5*delta;
                  max_value += 0.5*delta;
                 }
              delta = (max_value - min_value);
              if (delta == 0.) delta = 1.;
              value_norm = (value - min_value) / delta;
              if (value_norm > 1.) value_norm = 1.;
              if (value_norm < 0.) value_norm = 0.;
              valueBuf = value_norm * (pow(2.0,16.) -1.);
              outshort = (short)((float)valueBuf-32768);
              if (i < 10)
                 {printf("fr=%d  samp=%d  i=%d  value=%g  min=%g  max=%g  delta=%g  value_norm=%g  valueBuf=%g   outshort=%d\n",
                  frame->GTimeS, isamp, i, value, min_value, max_value, delta, value_norm, valueBuf, outshort);
                 }
             }

          isamp++;
          if (isamp == 1) previousShort = outshort;
          outshort0 = outshort;

          /*--- Upsample the data if needed ---*/
          for (z=1; z<=resampFactor; z++)
             {if (resampFactor > 1)
                 {valueBuf2 = ((resampFactor - z) * (double)previousShort + z * (double)outshort0) / resampFactor;
                  outshort = (short)((float)valueBuf2);
                 }
              fwrite(&outshort, sizeof(short), 1, fout);
              if (i < 10) printf("fr=%d  samp=%d  i+z=%d+%d  outshort=%d\n", frame->GTimeS, isamp, i, z, outshort);
             }
          previousShort = outshort;
         }

     FrameFree(frame);
     tt += framedt;
    }

 FrFileIEnd(iFile);
 fclose(fout);

 return;
}


/*-------------------------------------------------- DySoundReadHeader ------*/

void DySoundReadHeader (char *buffer)
{
 char format_tag_name[16];
 struct soundhdr myhdr;

 memcpy((char *)(&myhdr),buffer,sizeof(myhdr));

 strcpy(format_tag_name,"unknown");
 if (myhdr.format_tag == 1) strcpy(format_tag_name,"PCM");
 if (myhdr.format_tag == 257) strcpy(format_tag_name,"Mu-Law");
 if (myhdr.format_tag == 258) strcpy(format_tag_name,"A-Law");
 if (myhdr.format_tag == 259) strcpy(format_tag_name,"ADPCM");

 printf("Header size: %d bytes \n", sizeof(myhdr));
 printf("riff: %c  %c  %c  %c \n", myhdr.riff[0], myhdr.riff[1], myhdr.riff[2], myhdr.riff[3]);
 printf("      %x %x %x %x \n", buffer[0], buffer[1], buffer[2], buffer[3]);
 printf("File length: %ld bytes \n", myhdr.flength);
 printf("wave: %c%c%c%c \n", myhdr.wave[0], myhdr.wave[1], myhdr.wave[2], myhdr.wave[3]);
 printf("fmt: %c%c%c%c \n", myhdr.fmt[0], myhdr.fmt[1], myhdr.fmt[2], myhdr.fmt[3]);
 printf("Block size: %ld bytes \n", myhdr.block_size);
 printf("Format: %d (%s)  \n", myhdr.format_tag, format_tag_name);
 printf("Number of channels: %d  \n", myhdr.num_chans);
 printf("Sample Rate: %ld  \n", myhdr.srate);
 printf("Number of bytes per sec: %ld  \n", myhdr.bytes_per_sec);
 printf("Number of bytes per sample: %d  \n", myhdr.bytes_per_samp);
 printf("Number of bits per sample: %d  \n", myhdr.bits_per_samp);
 printf("data: %c%c%c%c \n", myhdr.data[0], myhdr.data[1], myhdr.data[2], myhdr.data[3]);
 printf("Data length: %ld  \n", myhdr.dlength);

/*
     header attendu (big-endian):
     R I  F F  file_length W A  V E  f m  t space
     5249 4646 2408 0000   5741 5645 666d 7420
     block_size format nch  sampRate    bytes_per_sec
     1000 0000  0100   0200 2256 0000   8858 0100
     byte_per_sample  bits_per_sample d a  t a  data_length
     0400             1000            6461 7461 0008 0000

     header obtenu sur lappc-vi22 avec un canal a 20kHz:
     4952 4646 58b4 0001 4157 4556 6d66 2074
     0010 0000 0001 0001 49d4 0000 93a8 0000
     0002 0010 6164 6174 5888 0001 a5a7 9a9f
*/

 return;
}


/*-------------------------------------------------- DySoundFillHeader ------*/

void DySoundFillHeader (struct soundhdr *hdr, long srate, long nBytes, short bits_per_sample)
{
 struct soundhdr myhdr = {"RIF",88244,"WAV","fmt",16,1,1,44100,88200,2,16,"dat",88200};
 myhdr.riff[3] = 'F';
 myhdr.wave[3] = 'E';
 myhdr.fmt[3] = ' ';
 myhdr.data[3] = 'a';
 myhdr.format_tag = 1;
 myhdr.num_chans = 1;
 myhdr.bits_per_samp = bits_per_sample;
 myhdr.bytes_per_samp = myhdr.num_chans * myhdr.bits_per_samp / 8;
 myhdr.srate = srate;
 myhdr.bytes_per_sec = srate * myhdr.bytes_per_samp;
 myhdr.dlength = nBytes;
 myhdr.flength = myhdr.dlength + sizeof(myhdr);
 *hdr = myhdr;

 return;
}


